"""Define MyQ devices."""
import asyncio
import logging
from datetime import datetime
from typing import TYPE_CHECKING, Optional, Union

from .device import MyQDevice
from .errors import RequestError

if TYPE_CHECKING:
    from .api import API

_LOGGER = logging.getLogger(__name__)

COMMAND_URI = \
    "https://account-devices-gdo.myq-cloud.com/api/v5.2/Accounts/{account_id}/door_openers/{device_serial}/{command}"
COMMAND_CLOSE = "close"
COMMAND_OPEN = "open"
STATE_CLOSED = "closed"
STATE_CLOSING = "closing"
STATE_OPEN = "open"
STATE_OPENING = "opening"
STATE_STOPPED = "stopped"
STATE_TRANSITION = "transition"
STATE_UNKNOWN = "unknown"


class MyQGaragedoor(MyQDevice):
    """Define a generic device."""

    def __init__(self, api: "API", device_json: dict, account: str, state_update: datetime) -> None:
        """Initialize.
        :type account: str
        """
        super().__init__(api=api, account=account, device_json=device_json, state_update=state_update)

    @property
    def close_allowed(self) -> bool:
        """Return whether the device can be closed unattended."""
        return self.device_json["state"].get("is_unattended_close_allowed") is True

    @property
    def open_allowed(self) -> bool:
        """Return whether the device can be opened unattended."""
        return self.device_json["state"].get("is_unattended_open_allowed") is True

    @property
    def device_state(self) -> Optional[str]:
        """Return the current state of the device."""
        return (
            self.device_json["state"].get("door_state")
            if self.device_json.get("state") is not None
            else None
        )

    async def close(self, wait_for_state: bool = False) -> Union[asyncio.Task, bool]:
        """Close the device."""
        if self.state != self.device_state:
            raise RequestError(f"Device is currently {self.state}, wait until complete.")

        if self.state not in (STATE_CLOSED, STATE_CLOSING):
            # If our state is different from device state then it means an action is already being performed.
            if self.state != self.device_state:
                raise RequestError(f"Device is currently {self.state}, wait until complete.")

            # Device is currently not closed or closing, send command to close
            await self._send_state_command(
                url=COMMAND_URI.format(
                    account_id=self.account,
                    device_serial=self.device_id,
                    command=COMMAND_CLOSE,
                ),
                command=COMMAND_CLOSE,
            )
            self.state = STATE_CLOSING

        wait_for_state_task = asyncio.create_task(self.wait_for_state(
            current_state=[STATE_CLOSING],
            new_state=[STATE_CLOSED],
            last_state_update=self.device_json["state"].get("last_update"),
            timeout=60,
        ), name="MyQ_WaitForClose",
        )
        if not wait_for_state:
            return wait_for_state_task

        _LOGGER.debug("Waiting till garage is closed")
        return await wait_for_state_task

    async def open(self, wait_for_state: bool = False) -> Union[asyncio.Task, bool]:
        """Open the device."""
        if self.state not in (STATE_OPEN, STATE_OPENING):
            # Set the current state to "opening" right away (in case the user doesn't
            # run update() before checking):
            await self._send_state_command(
                url=COMMAND_URI.format(
                    account_id=self.account,
                    device_serial=self.device_id,
                    command=COMMAND_OPEN,
                ),
                command=COMMAND_OPEN,
            )
            self.state = STATE_OPENING

        wait_for_state_task = asyncio.create_task(self.wait_for_state(
            current_state=[STATE_OPENING],
            new_state=[STATE_OPEN],
            last_state_update=self.device_json["state"].get("last_update"),
            timeout=60,
        ), name="MyQ_WaitForOpen",
        )

        if not wait_for_state:
            return wait_for_state_task

        _LOGGER.debug("Waiting till garage is open")
        return await wait_for_state_task



"""Define MyQ devices."""
import asyncio
import logging
from datetime import datetime
from typing import TYPE_CHECKING, Optional, Union

from .device import MyQDevice

if TYPE_CHECKING:
    from .api import API

_LOGGER = logging.getLogger(__name__)

COMMAND_URI = "https://account-devices-lamp.myq-cloud.com/api/v5.2/Accounts/{account_id}/lamps/{device_serial}/{command}"
COMMAND_ON = "on"
COMMAND_OFF = "off"
STATE_ON = "on"
STATE_OFF = "off"


class MyQLamp(MyQDevice):
    """Define a generic device."""

    def __init__(
        self, api: "API", device_json: dict, account: str, state_update: datetime
    ) -> None:
        """Initialize.
        :type account: str
        """
        super().__init__(
            api=api, account=account, device_json=device_json, state_update=state_update
        )

    @property
    def device_state(self) -> Optional[str]:
        """Return the current state of the device."""
        return (
            self.device_json["state"].get("lamp_state")
            if self.device_json.get("state") is not None
            else None
        )

    async def turnoff(self, wait_for_state: bool = False) -> Union[asyncio.Task, bool]:
        """Turn light off."""

        await self._send_state_command(
            url=COMMAND_URI.format(
                account_id=self.account,
                device_serial=self.device_id,
                command=COMMAND_OFF,
            ),
            command=COMMAND_OFF,
        )
        self.state = STATE_OFF

        wait_for_state_task = asyncio.create_task(
            self.wait_for_state(
                current_state=[STATE_ON],
                new_state=[STATE_OFF],
                last_state_update=self.device_json["state"].get("last_update"),
                timeout=30,
            ),
            name="MyQ_WaitForOff",
        )
        if not wait_for_state:
            return wait_for_state_task

        _LOGGER.debug("Waiting till light is off")
        return await wait_for_state_task

    async def turnon(self, wait_for_state: bool = False) -> Union[asyncio.Task, bool]:
        """Turn light on."""

        await self._send_state_command(
            url=COMMAND_URI.format(
                account_id=self.account,
                device_serial=self.device_id,
                command=COMMAND_ON,
            ),
            command=COMMAND_ON,
        )
        self.state = STATE_ON

        wait_for_state_task = asyncio.create_task(
            self.wait_for_state(
                current_state=[STATE_ON],
                new_state=[STATE_OFF],
                last_state_update=self.device_json["state"].get("last_update"),
                timeout=30,
            ),
            name="MyQ_WaitForOn",
        )
        if not wait_for_state:
            return wait_for_state_task

        _LOGGER.debug("Waiting till light is on")
        return await wait_for_state_task



"""Handle requests to MyQ."""
import asyncio
import logging
from json import JSONDecodeError

from aiohttp import ClientSession, ClientResponse
from aiohttp.client_exceptions import ClientError, ClientResponseError

from .errors import RequestError

_LOGGER = logging.getLogger(__name__)

REQUEST_METHODS = dict(
    json="request_json", text="request_text", response="request_response"
)
DEFAULT_REQUEST_RETRIES = 5


class MyQRequest:  # pylint: disable=too-many-instance-attributes
    """Define a class to handle requests to MyQ"""

    def __init__(self, websession: ClientSession = None, useragent: str = None) -> None:
        self._websession = websession or ClientSession()
        self._useragent = useragent

    async def _send_request(
        self,
        method: str,
        url: str,
        websession: ClientSession,
        headers: dict = None,
        params: dict = None,
        data: dict = None,
        json: dict = None,
        allow_redirects: bool = False,
    ) -> ClientResponse:

        attempt = 0
        resp_exc = None
        last_status = ""
        last_error = ""

        if self._useragent is not None:
            headers.update({"User-Agent": self._useragent})

        while attempt < DEFAULT_REQUEST_RETRIES:
            if attempt != 0:
                wait_for = min(2 ** attempt, 5)
                _LOGGER.debug(
                    f'Request failed with "{last_status} {last_error}" '
                    f'(attempt #{attempt}/{DEFAULT_REQUEST_RETRIES})"; trying again in {wait_for} seconds'
                )
                await asyncio.sleep(wait_for)

            attempt += 1
            try:
                _LOGGER.debug(
                    f"Sending myq api request {url} and headers {headers} with connection pooling"
                )
                resp = await websession.request(
                    method,
                    url,
                    headers=headers,
                    params=params,
                    data=data,
                    json=json,
                    skip_auto_headers={"USER-AGENT"},
                    allow_redirects=allow_redirects,
                    raise_for_status=True,
                )

                _LOGGER.debug("Response:")
                _LOGGER.debug(f"    Response Code: {resp.status}")
                _LOGGER.debug(f"    Headers: {resp.raw_headers}")
                _LOGGER.debug(f"    Body: {await resp.text()}")
                return resp
            except ClientResponseError as err:
                _LOGGER.debug(
                    f"Attempt {attempt} request failed with exception : {err.status} - {err.message}"
                )
                if err.status == 401:
                    raise err
                last_status = err.status
                last_error = err.message
                resp_exc = err
            except ClientError as err:
                _LOGGER.debug(
                    f"Attempt {attempt} request failed with exception:: {str(err)}"
                )
                last_status = ""
                last_error = str(err)
                resp_exc = err

        raise resp_exc

    async def request_json(
        self,
        method: str,
        url: str,
        websession: ClientSession = None,
        headers: dict = None,
        params: dict = None,
        data: dict = None,
        json: dict = None,
        allow_redirects: bool = False,
    ) -> (ClientResponse, dict):

        websession = websession or self._websession

        resp = await self._send_request(
            method=method,
            url=url,
            headers=headers,
            params=params,
            data=data,
            json=json,
            allow_redirects=allow_redirects,
            websession=websession,
        )

        try:
            data = await resp.json(content_type=None)
        except JSONDecodeError as err:
            message = (
                f"JSON Decoder error {err.msg} in response at line {err.lineno} column {err.colno}. Response "
                f"received was:\n{err.doc}"
            )
            _LOGGER.error(message)
            raise RequestError(message)

        return resp, data

    async def request_text(
        self,
        method: str,
        url: str,
        websession: ClientSession = None,
        headers: dict = None,
        params: dict = None,
        data: dict = None,
        json: dict = None,
        allow_redirects: bool = False,
    ) -> (ClientResponse, str):

        websession = websession or self._websession

        resp = await self._send_request(
            method=method,
            url=url,
            headers=headers,
            params=params,
            data=data,
            json=json,
            allow_redirects=allow_redirects,
            websession=websession,
        )

        return resp, await resp.text()

    async def request_response(
        self,
        method: str,
        url: str,
        websession: ClientSession = None,
        headers: dict = None,
        params: dict = None,
        data: dict = None,
        json: dict = None,
        allow_redirects: bool = False,
    ) -> (ClientResponse, None):

        websession = websession or self._websession

        return (
            await self._send_request(
                method=method,
                url=url,
                headers=headers,
                params=params,
                data=data,
                json=json,
                allow_redirects=allow_redirects,
                websession=websession,
            ),
            None,
        )



"""The myq constants."""

OAUTH_CLIENT_ID = "IOS_CGI_MYQ"
OAUTH_CLIENT_SECRET = "VUQ0RFhuS3lQV3EyNUJTdw=="
OAUTH_BASE_URI = "https://partner-identity.myq-cloud.com"
OAUTH_AUTHORIZE_URI = f"{OAUTH_BASE_URI}/connect/authorize"
OAUTH_REDIRECT_URI = "com.myqops://ios"
OAUTH_TOKEN_URI = f"{OAUTH_BASE_URI}/connect/token"

ACCOUNTS_ENDPOINT = "https://accounts.myq-cloud.com/api/v6.0/accounts"
DEVICES_ENDPOINT = "https://devices.myq-cloud.com/api/v5.2/Accounts/{account_id}/Devices"

WAIT_TIMEOUT = 60

DEVICE_TYPE = "device_type"
DEVICE_TYPE_GATE = "gate"
DEVICE_FAMILY = "device_family"
DEVICE_FAMILY_GATEWAY = "gateway"
DEVICE_FAMILY_GARAGEDOOR = "garagedoor"
DEVICE_FAMLY_LAMP = "lamp"
DEVICE_STATE = "state"
DEVICE_STATE_ONLINE = "online"

MANUFACTURER = "The Chamberlain Group Inc."

KNOWN_MODELS = {
    "00": "Chamberlain Ethernet Gateway",
    "01": "LiftMaster Ethernet Gateway",
    "02": "Craftsman Ethernet Gateway",
    "03": "Chamberlain Wi-Fi hub",
    "04": "LiftMaster Wi-Fi hub",
    "05": "Craftsman Wi-Fi hub",
    "08": "LiftMaster Wi-Fi GDO DC w/Battery Backup",
    "09": "Chamberlain Wi-Fi GDO DC w/Battery Backup",
    "10": "Craftsman Wi-Fi GDO DC 3/4HP",
    "11": "MyQ Replacement Logic Board Wi-Fi GDO DC 3/4HP",
    "12": "Chamberlain Wi-Fi GDO DC 1.25HP",
    "13": "LiftMaster Wi-Fi GDO DC 1.25HP",
    "14": "Craftsman Wi-Fi GDO DC 1.25HP",
    "15": "MyQ Replacement Logic Board Wi-Fi GDO DC 1.25HP",
    "0A": "Chamberlain Wi-Fi GDO or Gate Operator AC",
    "0B": "LiftMaster Wi-Fi GDO or Gate Operator AC",
    "0C": "Craftsman Wi-Fi GDO or Gate Operator AC",
    "0D": "MyQ Replacement Logic Board Wi-Fi GDO or Gate Operator AC",
    "0E": "Chamberlain Wi-Fi GDO DC 3/4HP",
    "0F": "LiftMaster Wi-Fi GDO DC 3/4HP",
    "20": "Chamberlain MyQ Home Bridge",
    "21": "LiftMaster MyQ Home Bridge",
    "23": "Chamberlain Smart Garage Hub",
    "24": "LiftMaster Smart Garage Hub",
    "27": "LiftMaster Wi-Fi Wall Mount opener",
    "28": "LiftMaster Commercial Wi-Fi Wall Mount operator",
    "80": "EU LiftMaster Ethernet Gateway",
    "81": "EU Chamberlain Ethernet Gateway",
}
